""" Add new sports

Revision ID: cee0830497f8
Revises: 4e8597c50064
Create Date: 2021-08-25 13:58:52.333603

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = 'cee0830497f8'
down_revision = '4e8597c50064'
branch_labels = None
depends_on = None


def upgrade():

    op.execute(
        """
        INSERT INTO sports (label, img, is_active)
        VALUES 
            ('Mountain Biking (Electric)', '/img/sports/electric-mountain-biking.png', True),
            ('Trail', '/img/sports/trail.png', True),
            ('Skiing (Alpine)', '/img/sports/alpine-skiing.png', True),
            ('Skiing (Cross Country)', '/img/sports/cross-country-skiing.png', True),
            ('Rowing', '/img/sports/rowing.png', True)
        """
    )


def downgrade():
    op.execute(
        """
        DELETE FROM sports
        WHERE label = 'Mountain Biking (Electric)'
           OR label = 'Trail'
           OR label = 'Skiing (Alpine)'
           OR label = 'Skiing (Cross Country)'
           OR label = 'Rowing';
        """
    )



from abc import ABC, abstractmethod
from datetime import datetime
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import pytz

from fittrackee.equipments.utils import handle_equipments
from fittrackee.users.models import UserSportPreference
from fittrackee.workouts.models import Sport

from ..exceptions import WorkoutException
from ..models import TITLE_MAX_CHARACTERS, Workout

if TYPE_CHECKING:
    from fittrackee.equipments.models import Equipment
    from fittrackee.users.models import User
    from fittrackee.workouts.models import Workout


class BaseWorkoutService(ABC):
    def __init__(
        self,
        auth_user: "User",
        sport_id: Optional[int],
        equipment_ids: Union[List[str], None],
    ):
        if not sport_id:
            raise WorkoutException("invalid", "no sport id provided")
        try:
            self.sport = Sport.query.filter_by(id=sport_id).one()
        except Exception as e:
            raise WorkoutException(
                "invalid",
                f"Sport id: {sport_id} does not exist",
            ) from e
        self.auth_user = auth_user
        self.sport_preferences = UserSportPreference.query.filter_by(
            user_id=self.auth_user.id, sport_id=self.sport.id
        ).first()
        self.stopped_speed_threshold = (
            self.sport.stopped_speed_threshold
            if self.sport_preferences is None
            else self.sport_preferences.stopped_speed_threshold
        )
        self.equipment_ids = equipment_ids

    def get_equipments(self) -> Union[List["Equipment"], None]:
        if self.equipment_ids is None and self.sport_preferences:
            return [
                equipment
                for equipment in self.sport_preferences.default_equipments.all()  # noqa
                if equipment.is_active is True
            ]
        return handle_equipments(
            self.equipment_ids,
            self.auth_user,
            self.sport.id,
        )

    def _get_title(self, workout_date: datetime, title: Optional[str]) -> str:
        if title:
            return title[:TITLE_MAX_CHARACTERS]

        workout_datetime = (
            workout_date.astimezone(pytz.timezone(self.auth_user.timezone))
            if self.auth_user.timezone
            else workout_date
        ).strftime("%Y-%m-%d %H:%M:%S")
        return f"{self.sport.label} - {workout_datetime}"

    @abstractmethod
    def process(self) -> Tuple[List["Workout"], Dict]:
        """
        returns:
        - list of workouts created successfully
        - a dict with errored workouts and flag indicating asynchronous
          processing
        """
        pass



"""Add date_format for date display to user preferences in DB

Revision ID: bf13b8f5589d
Revises: 84d840ce853b
Create Date: 2022-10-25 18:53:59.378423

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'bf13b8f5589d'
down_revision = '5b936821326d'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        'users', sa.Column('date_format', sa.String(length=50), nullable=True)
    )
    op.execute("UPDATE users SET date_format = 'MM/dd/yyyy'")
    op.alter_column('users', 'date_format', nullable=False)



def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('users', 'date_format')
    # ### end Alembic commands ###



from typing import Dict, List

from flask import Flask

from fittrackee import db

from ..dates import get_datetime_in_utc
from .models import AppConfig

MAX_FILE_SIZE = 1 * 1024 * 1024  # 1MB


def get_or_init_config() -> AppConfig:
    """
    Init application configuration.
    """
    existing_config = AppConfig.query.one_or_none()
    if existing_config:
        return existing_config
    config = AppConfig()
    config.max_users = 0  # no limitation
    config.max_single_file_size = MAX_FILE_SIZE
    config.max_zip_file_size = MAX_FILE_SIZE * 10
    db.session.add(config)
    db.session.commit()
    return config


def update_app_config_from_database(
    current_app: Flask, db_config: AppConfig
) -> None:
    current_app.config.update(
        {
            "file_limit_import": db_config.file_limit_import,
            "file_sync_limit_import": db_config.file_sync_limit_import,
            "max_single_file_size": db_config.max_single_file_size,
            "MAX_CONTENT_LENGTH": db_config.max_zip_file_size,
            "max_users": db_config.max_users,
            "is_registration_enabled": db_config.is_registration_enabled,
            "privacy_policy_date": (
                db_config.privacy_policy_date
                if db_config.privacy_policy
                else get_datetime_in_utc(
                    current_app.config["DEFAULT_PRIVACY_POLICY_DATA"],
                    "%Y-%m-%d %H:%M:%S",
                )
            ),
            "stats_workouts_limit": db_config.stats_workouts_limit,
        }
    )


def verify_app_config(config_data: Dict) -> List:
    """
    Verify if application config is valid.

    If not, it returns not empty string
    """
    ret = []

    params = [
        (
            "file_sync_limit_import",
            "max files in a zip archive processed synchronously",
        ),
        ("file_limit_import", "max files in a zip archive"),
        ("max_single_file_size", "max size of uploaded files"),
        ("max_zip_file_size", "max size of zip archive"),
    ]
    for param, label in params:
        if param in config_data and config_data[param] <= 0:
            ret.append(f"{label} must be greater than 0")

    params = [
        ("max_users", "max users"),
        ("stats_workouts_limit", "max number of workouts for statistics"),
    ]
    for param, label in params:
        if param in config_data and config_data[param] < 0:
            ret.append(f"{label} must be greater than or equal to 0")

    return ret



from typing import List, Optional, Union

from fittrackee.equipments.models import Equipment
from fittrackee.users.models import User
from fittrackee.utils import decode_short_id
from fittrackee.workouts.models import Sport

from .exceptions import InvalidEquipmentException, InvalidEquipmentsException

SPORT_EQUIPMENT_TYPES = {
    "Shoes": [
        "Hiking",
        "Mountaineering",
        "Running",
        "Swimrun",
        "Trail",
        "Walking",
    ],
    "Bike": [
        "Cycling (Sport)",
        "Cycling (Transport)",
        "Cycling (Trekking)",
        "Halfbike",
        "Mountain Biking",
        "Mountain Biking (Electric)",
    ],
    "Bike Trainer": ["Cycling (Virtual)"],
    "Board": ["Windsurfing"],
    "Kayak_Boat": ["Canoeing", "Rowing", "Kayaking"],
    "Skis": ["Skiing (Alpine)", "Skiing (Cross Country)"],
    "Snowshoes": ["Snowshoes"],
}


def handle_equipments(
    equipment_short_ids: Optional[List[str]],
    auth_user: User,
    sport_id: int,
    existing_equipments: Optional[List[Equipment]] = None,
) -> Union[List[Equipment], None]:
    equipments_list: Optional[List[Equipment]] = None

    if equipment_short_ids is not None:
        sport = Sport.query.filter_by(id=sport_id).first()
        if not sport:
            raise InvalidEquipmentsException(f"sport id {sport_id} not found")

        equipments_list = []
        if not isinstance(equipment_short_ids, list):
            raise InvalidEquipmentsException(
                "equipment_ids must be an array of strings"
            )

        # for now only one equipment par sport or workout
        if len(equipment_short_ids) > 1:
            raise InvalidEquipmentsException("only one equipment can be added")

        for equipment_short_id in equipment_short_ids:
            if not isinstance(equipment_short_id, str):
                raise InvalidEquipmentsException(
                    "equipment_ids must be an array of strings"
                )
            equipment = Equipment.query.filter_by(
                uuid=decode_short_id(equipment_short_id), user_id=auth_user.id
            ).first()
            if not equipment:
                raise InvalidEquipmentException(
                    status="not_found",
                    message=(
                        f"equipment with id {equipment_short_id} "
                        "does not exist"
                    ),
                    equipment_short_id=equipment_short_id,
                )

            if sport.label not in SPORT_EQUIPMENT_TYPES.get(
                equipment.equipment_type.label, []
            ):
                raise InvalidEquipmentException(
                    status="invalid",
                    message=(
                        f"invalid equipment id {equipment.short_id} "
                        f"for sport {sport.label}"
                    ),
                    equipment_short_id=equipment_short_id,
                )

            if not equipment.is_active and (
                not existing_equipments or equipment not in existing_equipments
            ):
                raise InvalidEquipmentException(
                    status="inactive",
                    message=(
                        f"equipment with id {equipment_short_id} is inactive"
                    ),
                    equipment_short_id=equipment_short_id,
                )
            equipments_list.append(equipment)
    return equipments_list



from functools import wraps
from typing import Any, Callable, List, Union

from authlib.integrations.flask_oauth2 import ResourceProtector
from authlib.oauth2 import OAuth2Error
from authlib.oauth2.rfc6749.errors import MissingAuthorizationError
from flask import current_app, request
from werkzeug.exceptions import RequestEntityTooLarge

from fittrackee.responses import (
    ForbiddenErrorResponse,
    PayloadTooLargeErrorResponse,
    UnauthorizedErrorResponse,
)
from fittrackee.users.models import User
from fittrackee.users.roles import UserRole


class CustomResourceProtector(ResourceProtector):
    def __call__(
        self,
        scopes: Union[str, List, None] = None,
        optional: bool = False,
        *,
        role: Union[UserRole, None] = None,
        optional_auth_user: bool = False,
        allow_suspended_user: bool = False,
    ) -> Callable:
        def wrapper(f: Callable) -> Callable:
            @wraps(f)
            def decorated(*args: Any, **kwargs: Any) -> Callable:
                auth_user = None
                auth_header = request.headers.get("Authorization")
                if not optional_auth_user and not auth_header:
                    return UnauthorizedErrorResponse(
                        "provide a valid auth token"
                    )

                if auth_header:
                    # First-party application (Fittrackee front-end)
                    # in this case, scopes will be ignored
                    auth_token = auth_header.split(" ")[1]
                    resp = User.decode_auth_token(auth_token)
                    if isinstance(resp, int):
                        auth_user = User.query.filter_by(id=resp).first()

                    # Third-party applications
                    if not auth_user and scopes:
                        current_token = None
                        try:
                            current_token = self.acquire_token(scopes)
                        except (
                            MissingAuthorizationError,
                            OAuth2Error,
                        ) as error:
                            self.raise_error_response(error)
                        except RequestEntityTooLarge:
                            file_type = ""
                            if request.endpoint in [
                                "auth.edit_picture",
                                "workouts.post_workout",
                            ]:
                                file_type = (
                                    "picture"
                                    if request.endpoint == "auth.edit_picture"
                                    else "workout"
                                )
                            return PayloadTooLargeErrorResponse(
                                file_type=file_type,
                                file_size=request.content_length,
                                max_size=current_app.config[
                                    "MAX_CONTENT_LENGTH"
                                ],
                            )
                        auth_user = (
                            None
                            if current_token is None
                            else current_token.user
                        )

                if (not optional_auth_user and not auth_user) or (
                    auth_user and not auth_user.is_active
                ):
                    return UnauthorizedErrorResponse(
                        "provide a valid auth token"
                    )

                if auth_user and (
                    (
                        allow_suspended_user is False
                        and auth_user.suspended_at is not None
                    )
                    or (role and auth_user.role < role.value)
                ):
                    return ForbiddenErrorResponse(
                        "you do not have permissions"
                        + (
                            ", your account is suspended"
                            if auth_user and auth_user.suspended_at
                            else ""
                        )
                    )
                return f(auth_user, *args, **kwargs)

            return decorated

        return wrapper



import hashlib
import random
from abc import abstractmethod
from typing import IO, TYPE_CHECKING, Dict, List, Optional, Union

from flask import current_app
from staticmap3 import Line, StaticMap

from fittrackee import VERSION, db
from fittrackee.files import get_absolute_file_path

from ..weather import WeatherService
from .workout_point import WorkoutPoint

if TYPE_CHECKING:
    from datetime import datetime

    from gpxpy.gpx import GPX

    from fittrackee.users.models import User
    from fittrackee.workouts.models import Workout

weather_service = WeatherService()


class BaseWorkoutWithSegmentsCreationService:
    # file is converted in gpx format if not in gpx format
    gpx: "GPX"

    @abstractmethod
    def __init__(
        self,
        auth_user: "User",
        workout_file: IO[bytes],
        sport_id: int,
        stopped_speed_threshold: float,
        get_weather: bool = True,
    ) -> None:
        self.auth_user = auth_user
        self.sport_id = sport_id
        self.coordinates: List[List[float]] = []
        self.stopped_speed_threshold = stopped_speed_threshold
        self.workout_name: Optional[str] = None
        self.workout_description: Optional[str] = None
        self.start_point: Optional["WorkoutPoint"] = None
        self.end_point: Optional["WorkoutPoint"] = None
        self.get_weather = get_weather

    @abstractmethod
    def get_workout_date(self) -> "datetime":
        pass

    @classmethod
    def get_map_hash(cls, map_filepath: str) -> str:
        """
        Generate a md5 hash used as id instead of workout id, to retrieve map
        image (maps are sensitive data)
        """
        md5 = hashlib.md5(usedforsecurity=False)
        absolute_map_filepath = get_absolute_file_path(map_filepath)
        with open(absolute_map_filepath, "rb") as f:
            for chunk in iter(lambda: f.read(128 * md5.block_size), b""):
                md5.update(chunk)
        return md5.hexdigest()

    @classmethod
    def get_static_map_tile_server_url(cls, tile_server_config: Dict) -> str:
        if tile_server_config["STATICMAP_SUBDOMAINS"]:
            subdomains = tile_server_config["STATICMAP_SUBDOMAINS"].split(",")
            subdomain = f"{random.choice(subdomains)}."  # nosec
        else:
            subdomain = ""
        return tile_server_config["URL"].replace("{s}.", subdomain)

    @classmethod
    def generate_map_image(cls, map_filepath: str, coordinates: List) -> None:
        m = StaticMap(
            width=400,
            height=225,
            padding_x=10,
            headers={"User-Agent": f"FitTrackee v{VERSION}"},
            delay_between_retries=5,
        )
        if not current_app.config["TILE_SERVER"]["DEFAULT_STATICMAP"]:
            m.url_template = cls.get_static_map_tile_server_url(
                current_app.config["TILE_SERVER"]
            )
        line = Line(coords=coordinates, color="#3388FF", width=4)
        m.add_line(line)
        image = m.render()
        image.save(map_filepath)

    @staticmethod
    def get_weather_data(
        start_point: WorkoutPoint, end_point: WorkoutPoint
    ) -> List[Union[Dict, None]]:
        return [
            weather_service.get_weather(start_point),
            weather_service.get_weather(end_point),
        ]

    @abstractmethod
    def _process_file(self) -> "Workout":
        pass

    def process_workout(self) -> "Workout":
        new_workout = self._process_file()

        if self.get_weather and self.start_point and self.end_point:
            new_workout.weather_start, new_workout.weather_end = (
                self.get_weather_data(
                    self.start_point,
                    self.end_point,
                )
            )

        db.session.flush()
        return new_workout
